import difflib
import json
import re

RED_COLOR = '\033[91m\b'
GREEN_COLOR = '\033[92m\b'
RESET_COLOR = '\033[0m'


def diff_strings(a, b, colorize=True):
    difference = difflib.Differ()

    lines = difference.compare(a.splitlines(), b.splitlines())
    lines = [colorize_diff_line(line, colorize) for line in lines]

    return "\n".join(lines)


def colorize_diff_line(line, colorize):
    if not colorize:
        return line

    line_color = ""
    if line.startswith("?"):
        if "+" in line:
            line_color = GREEN_COLOR
        elif "-" in line:
            line_color = RED_COLOR
    elif line.startswith("-"):
        line_color = RED_COLOR
    return line_color + line + RESET_COLOR


def colorize(s, color):
    return color + s + RESET_COLOR


def tojsonstr(o):
    return json.dumps(o, sort_keys=True, indent=2)


def coerce_str_to_int(s):
    try:
        return int(s)
    except ValueError:
        return s


def matches_any(s, patterns):
    for p in patterns:
        if p.match(s):
            return True
    return False


def compile_regexps(regexps):
    return [re.compile(".*" + p) for p in regexps]



import codecs
import json
import os
from pathlib import Path
from contextlib import contextmanager
from importlib.resources import files

_tmp_files = []


def list_files(path, include_ext):
    result = []
    for root, subdirs, files in os.walk(path):
        subdirs.sort()
        for filename in sorted(files):
            if filename.endswith(include_ext):
                result.append(os.path.join(root, filename))

    return result


def parse_json(filename):
    with codecs.open(filename, 'r', encoding='utf8') as f:
        return json.load(f)


def write_tmp(filename, content):
    with open(filename, "w") as fp:
        fp.write(str(content))
        _tmp_files.append(filename)


def cleanup_tmp_files():
    for filename in _tmp_files:
        os.remove(filename)


def read_file_contents(filename):
    with open(filename) as fp:
        return fp.read()




# coding=utf-8
import importlib.util
import inspect
import logging
import os.path

from . import matchers
from .util import file_util, log_util

_logger = log_util.get_logger(__name__, level=logging.DEBUG)


def load(conf):
    custom_matchers_dir = conf.get("matchers")
    if custom_matchers_dir:
        for file in file_util.list_files(custom_matchers_dir, ".py"):
            _logger.info("* Loading custom matcher: %s" % file)
            custom_matcher = CustomMatcher(file)
            # NOTE: This function has the side-effect of affect the matcher module
            matchers.add_matcher(custom_matcher.matcher_name, custom_matcher.match)


class CustomMatcher(object):
    def __init__(self, source_file):
        # this is constructor loads a plugin aka a custom matcher, the contract between us
        # and the user is that we expect that:
        # * the file contains a function called "match"
        # * the function takes two parameters: expected & actual
        # * expected is a string, actual is json (what we get back from the server)
        # * the function CAN return just true if the matcher passed, false otherwise
        # * the function is RECOMMENDED to return a tuple (boolean, string)
        # * if the function returns a tuple, the string will be shown as the message when the matcher failes
        try:
            self.matcher_func_name = "match"
            self.matcher_name = os.path.basename(source_file).split(".")[0]
            spec = importlib.util.spec_from_file_location(self.matcher_name, source_file)
            self.matcher_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(self.matcher_module)
            self.matcher_func = self.matcher_module.__getattribute__(self.matcher_func_name)
            CustomMatcher.validate_matcher(self.matcher_func)
        except Exception as e:
            raise AssertionError("Failed to load matcher %s: %s" % (source_file, e))

    @staticmethod
    def validate_matcher(matcher_func):
        if matcher_func is None:
            raise AssertionError("Expected to find 'match' function")
        arguments = inspect.getfullargspec(matcher_func).args
        expected_signature = ["expected", "actual"]
        if not arguments == expected_signature:
            raise AssertionError(
                "Expected 'match' to take exactly 2 parameters: %s - but was %s" % (expected_signature, arguments))

    def match(self, expected, actual):
        try:
            result = self.matcher_func(expected.strip(), actual)
            if isinstance(result, (bool)):
                return result, ""
            elif isinstance(result, (tuple)):
                return result
            else:
                raise AssertionError("Unexpected result %s from %s" % (result, self.matcher_name))
        except Exception as e:
            raise Exception("Custom matcher threw unexpected execption: %s" % str(e))



import requests

from .str_util import tojsonstr

_methods = ["get", "post", "put", "patch", "delete"]
_session = requests.Session()


def do_request(url, method, data, upload_file, headers, logger):
    http_verb = getattr(_session, method)
    logger.debug("--- REQUEST ---")
    logger.debug("%s: %s" % (method.upper(), url))
    logger.debug(" data: %s" % data)
    logger.debug(" headers: %s" % headers)
    logger.debug("----------------")

    # it's not possible to both upload a file & provide json data
    if upload_file:
        r = http_verb(url, files=upload_file, headers=headers)
    else:
        r = http_verb(url, data=data, headers=headers)

    logger.debug("--- RESPONSE ---")
    logger.debug("%s" % r.status_code)
    logger.debug("%s" % tojsonstr(as_json(r)))
    logger.debug("----------------")

    return r


def as_json(response):
    try:
        return response.json()
    except:
        return None



import collections.abc

def subset(d, keys):
    return {k: d[k] for k in keys if k in d}


def map_nested_dicts_py(d, func):
    if isinstance(d, collections.abc.Mapping):
        return {k: map_nested_dicts_py(v, func) for k, v in d.items()}
    else:
        return func(d)



"""Built-in matchers used by skivvy"""
# coding=utf-8
import string
import re
import os.path
from datetime import datetime
from math import fabs

import requests

from .util.str_util import coerce_str_to_int
from .util import file_util
from .util import log_util

_logger = log_util.get_logger(__name__)

DEFAULT_APPROXIMATE_THRESHOLD = 0.05  # default margin of error for a ~value to still be considered equal to another
SUCCESS_MSG = "OK"


def strip_matcher_prefix(s):
    if s.startswith("$"):
        return s[1:]
    else:
        return s


def match_expression(expected, actual):
    result = eval(expected, {}, {"actual": actual})
    if result is True:
        return True, SUCCESS_MSG
    else:
        return False, "Expected '%s' to evaluate to True but was evaluated to False" % actual


def match_regexp(expected, actual):
    try:
        expected = expected.strip()
        _logger.debug("Comparing '%s' to regexp: '%s'" % (actual, expected))
        if re.match(expected, actual):
            _logger.debug("It's a match.")
            return True, SUCCESS_MSG
        else:
            return False, "Expected '%s' to match regular expression '%s' - but didn't" % (actual, expected)
    except:
        return False, "Invalid regular expression in testcase: %s" % expected


def match_valid_url(expected, actual):
    try:
        # we allow relative urls using the format "$valid_url /some_url prefix example.com" -> example.com/some_url
        relative_url = expected.split("prefix")
        if len(relative_url) == 2:
            prefix = relative_url[-1]
            prefix = prefix.strip()
            actual = prefix + actual

        valid_status_codes = [200]
        _logger.debug("Making request to %s" % actual)
        response = requests.get(actual, verify=False)
        if response.status_code in valid_status_codes:
            _logger.debug("Success.")
            return True, SUCCESS_MSG
        else:
            _logger.debug("Failure.")
            return False, "Expected %s but got %s" % (valid_status_codes, actual)
    except Exception as e:
        _logger.debug("Failure.")
        _logger.debug("http call failed for: %s" % actual)
        _logger.debug("expected: %s" % expected)
        return False, "Failed to make request to %s: %s" % (actual, e)


def match_text(expected, actual):
    if not actual:
        return False, "Expected %s but got %s" % (expected, actual)

    for c in actual:
        if c not in string.ascii_letters:
            return False, "Expected %s but got %s" % (expected, actual)

    return True, SUCCESS_MSG


def match_contains(expected, actual):
    expected = expected.strip()
    actual = str(actual)

    if expected in actual:
        return True, SUCCESS_MSG
    else:
        return False, "Expected %s but got %s" % (expected, actual)


def default_matcher(expected, actual):
    if expected == actual:
        return True, SUCCESS_MSG
    else:
        return False, "Expected %s but got %s" % (expected, actual)


def is_almost_equal(expected, actual, threshold):
    abs_threshold = fabs(expected * threshold)
    delta = fabs(expected - actual)
    if delta < abs_threshold:
        return True, SUCCESS_MSG
    else:
        return False, "Expected %s±%s but get %s" % (expected, abs_threshold, actual)


def parse_threshold(expected):
    if "threshold" not in expected:
        return DEFAULT_APPROXIMATE_THRESHOLD, expected

    parts = expected.split("threshold")
    threshold = _parse_single_number(parts[1])
    return threshold, parts[0]


def len_match(expected, actual):
    try:
        len(actual)
    except Exception as e:
        return False, str(e)

    expected = expected.strip()
    threshold, expected = parse_threshold(expected)
    expected_value = _parse_single_number(expected)
    actual_value = len(actual)

    if expected.startswith("~"):
        return is_almost_equal(expected_value, actual_value, threshold)

    return default_matcher(expected_value, actual_value)


def len_greater_match(expected, actual):
    try:
        len(actual)
    except Exception as e:
        return False, str(e)

    expected_value = _parse_single_number(expected.strip())
    actual_value = len(actual)

    return actual_value > expected_value, "Expected %s>%s" % (actual_value, expected_value)


def len_less_match(expected, actual):
    try:
        len(actual)
    except Exception as e:
        return False, str(e)

    expected_value = _parse_single_number(expected.strip())
    actual_value = len(actual)

    return actual_value < expected_value, "Expected %s<%s" % (actual_value, expected_value)


def approximate_match(expected, actual):
    expected = expected.strip()
    threshold, expected = parse_threshold(expected)
    expected_value = _parse_single_number(expected)
    actual = float(actual)
    return is_almost_equal(expected_value, actual, threshold)


def date_matcher(expected, actual):
    expected = expected.strip()

    if expected == "today":
        expected = datetime.today().date()
        actual = datetime.strptime(actual[0:10], "%Y-%m-%d").date()
    else:
        return False, "DATE FORMAT '%s' NOT SUPPORTED!" % expected

    if actual == expected:
        return True, SUCCESS_MSG
    else:
        return False, "Expected %s but got %s" % (expected, actual)


def match_valid_ip(expected, actual):
    if "." in actual:
        return _match_valid_ip4(actual)
    else:
        return _match_valid_ip6(actual)


def _match_valid_ip4(actual):
    parts = actual.split(".")
    if len(parts) != 4:
        return False, "Expected valid ip but got %s" % actual
    for part in parts:
        part = int(part)
        if part < 0 or part > 255:
            return False, "Expected valid ip but got %s" % actual
    return True, ""


def _match_valid_ip6(actual):
    parts = actual.split(":")
    if len(parts) != 7:
        return False, "Expected valid ip but got %s" % actual
    for part in parts:
        if len(part) != 4:
            return False, "Expected valid ip but got %s" % actual
    return True, ""


def file_writer(expected, actual):
    expected = expected.strip()
    file_util.write_tmp(expected, actual)
    return True, SUCCESS_MSG


def file_reader(expected, actual):
    expected = expected.strip()
    data = file_util.read_file_contents(expected)
    data = data.strip()
    is_match = str(data) == str(actual)
    error_msg = "Files content in %s didn't match - expected: %s but got %s" % (expected, data, actual)
    if not is_match:
        _logger.warning(error_msg)
    return is_match, error_msg


def _parse_single_number(expected):
    # skip characters in the beginning which aren't digits
    index = 0
    for c in expected:
        if c.isdigit() or c == "-":
            break
        index += 1

    return float(expected[index:])


# when brace expansion is not used, just return the same string as passed in
def brace_expand_noop(s):
    return s


# technically not a matcher but this file seems like the best location nonetheless?
def brace_expand(s, auto_coerce):
    if not isinstance(s, str):
        return s

    while True:
        match = re.search(r"\<.*?\>", s)
        if not match:
            break
        else:
            variable = match.group(0)
            variable_name = variable.replace("<", "").replace(">", "")
            # maybe it refers to a file?
            if not os.path.isfile(variable_name):
                # Oh well, I give up! ;)
                _logger.warning("Failed to match variable %s to anything" % variable)
                break
            else:
                variable_value = file_util.read_file_contents(variable_name)
                s = s.replace(variable, variable_value)

    if auto_coerce:
        return coerce_str_to_int(s)
    else:
        return s


def add_matcher(matcher_name, matcher_func):
    if matcher_name in matcher_dict:
        raise AssertionError("Duplicate matcher: %s" % matcher_name)
    matcher_dict["$" + matcher_name] = matcher_func


def negating_matcher(negating_name, matcher_func):
    def do_match(expected, actual):
        result, msg = matcher_func(expected, actual)
        if result:
            return False, "Expected negating matcher ('$!%s %s') - to be FALSE but was TRUE for: %s" % (
                negating_name, expected, actual)
        else:
            return True, None

    return do_match


def add_negating_matchers():
    negating_matchers = []
    for matcher_name, matcher_func in matcher_dict.items():
        matcher_name = strip_matcher_prefix(matcher_name)
        negating_matchers.append(("!"+matcher_name, negating_matcher(matcher_name, matcher_func)))

    for name, func in negating_matchers:
        add_matcher(name, func)


matcher_dict = {
    "$valid_url": match_valid_url,
    "$contains": match_contains,
    "$len": len_match,
    "$len_gt": len_greater_match,
    "$len_lt": len_less_match,
    "$~": approximate_match,
    "$date": date_matcher,
    "$write_file": file_writer,
    "$read_file": file_reader,
    "$valid_ip": match_valid_ip,
    "$expects": match_expression,
    "$text": match_text,
    "$regexp": match_regexp,
    "$expr": match_expression,
}



from . import matchers
from .util import log_util
from .util.str_util import tojsonstr

_logger = log_util.get_logger(__name__)

# basically just looks at a string and the name of a matcher and determines if the string is invoking that matcher
# returns true in that case, false otherwise
def has_matcher_syntax(expected, matcher):
    try:
        first_word = expected.split()[0]
        if isinstance(expected, str) and first_word == matcher:
            return True
        return False
    except:
        return False


def is_matcher(expected):
    for matcher in matchers.matcher_dict.keys():
        if has_matcher_syntax(expected, matcher):
            return True
    return False


def verify_dict(expected, actual, **match_options):
    for key in expected.keys():
        _logger.debug("Checking '%s'..." % key)
        verify(expected.get(key), actual.get(key), **match_options)
        _logger.debug("Success.")


# TODO: we should support strict or non-strict types of comparisons of lists
def verify_list(expected, actual, **match_options):
    match_subsets = match_options.get("match_subsets", False)

    for expected_entry in expected:
        _logger.debug("Checking '%s'..." % expected_entry)
        if expected_entry not in actual:
            if match_subsets:
                verify_list_subset(expected_entry, actual, **match_options)
            else:
                raise Exception("Didn't find:\n%s\nin:\n%s" % (tojsonstr(expected_entry), tojsonstr(actual)))


def verify_list_subset(expected_entry, actual, **match_options):
    if isinstance(actual, list):
        for actual_entry in actual:
            if isinstance(actual_entry, dict) and isinstance(expected_entry, dict):
                e = dict(expected_entry)
                a = dict(actual_entry)

                a.update(e)

                try:
                    verify(a, actual_entry, **match_options)
                    return
                except:
                    pass

    raise Exception("Didn't find:\n%s\nin:\n%s" % (tojsonstr(expected_entry), tojsonstr(actual)))


def verify_matcher(expected, actual):
    for matcher in matchers.matcher_dict.keys():
        if has_matcher_syntax(expected, matcher):
            expected = expected[len(matcher):]
            matcher_func = matchers.matcher_dict.get(matcher)
            result, msg = matcher_func(expected, actual)
            if not result:
                raise Exception(msg)

    return matchers.default_matcher(expected, actual)


def verify(expected, actual, **match_options):
    if is_matcher(expected):
        verify_matcher(expected, actual)
    elif type(expected) != type(actual):
        if not actual and not expected:
            if match_options.get("match_falsiness"):
                return True
        raise Exception("%s is not the same type as %s" % (expected, actual))
    elif isinstance(expected, dict):
        return verify_dict(expected, actual, **match_options)
    elif isinstance(expected, list):
        return verify_list(expected, actual, **match_options)
    elif expected != actual:
        raise Exception("expected %s but was %s" % (expected, actual))
    else:
        return True



"""skivvy

Usage:
    skivvy.py run <cfg_file>
    skivvy.py run <cfg_file> [-t]
    skivvy.py run <cfg_file> [-t] -i path.*file...
    skivvy.py run <cfg_file> [-t] -e path.*file...
    skivvy.py run <cfg_file> [-t] -i path.*file... -e path.*file...

    skivvy.py run cfg/example.json (run examples)

Options:
  -h --help         Show this screen.
  -v --version      Show version.
  -i=regexp         Only include files matching any provided regexp(s) [default: .*]
  -e=regexp         Exclude files matching provided regexp(s)
  -t                Keep temporary files (if any)
"""
import json
import logging
from functools import partial
from urllib.parse import urljoin

from docopt import docopt

from . import custom_matchers
from . import matchers
from .skivvy_config import read_config
from .util import file_util, http_util, dict_util, str_util
from .util import log_util
from .util.str_util import tojsonstr, diff_strings, RED_COLOR
from .verify import verify

version = "0.510"
STATUS_OK = "OK"
STATUS_FAILED = "FAILED"

_logger = log_util.get_logger(__name__, level=logging.DEBUG)


def configure_testcase(test_dict, conf_dict):
    testcase = dict(conf_dict)
    testcase.update(test_dict)
    return testcase


def configure_logging(testcase):
    log_level = testcase.get("log_level", None)
    if log_level:
        _logger.setLevel(log_level)


def override_default_headers(default_headers, more_headers):
    d = dict(default_headers)
    d.update(more_headers)
    return d


def deprecation_warnings(testcase):
    if "url_brace_expansion" in testcase:
        print("Warning - 'url_brace_expansion' has been deprecated: use 'brace_expansion' instead.")


def log_testcase_failed(testfile, conf):
    colorize = conf.get("colorize", True)
    failure_msg = "\n\n%s\t%s\n\n" % (testfile, STATUS_FAILED)
    if colorize:
        failure_msg = str_util.colorize(failure_msg, RED_COLOR)
    _logger.error(failure_msg)


def log_error_context(err_context, conf):
    colorize = conf.get("colorize", True)
    e, expected, actual = err_context["exception"], err_context["expected"], err_context["actual"]
    _logger.error(str(e))
    _logger.info("--------------- DIFF BEGIN ---------------")
    diff_output = diff_strings(tojsonstr(expected), tojsonstr(actual), colorize=colorize)
    _logger.info(diff_output)
    _logger.info("--------------- DIFF END -----------------")
    _logger.debug("************** EXPECTED *****************")
    _logger.debug("!!! expected:\n%s" % tojsonstr(expected))
    _logger.debug("**************  ACTUAL *****************")
    _logger.debug("!!! actual:\n%s" % tojsonstr(actual))
    _logger.debug("\n" * 5)


def run_test(filename, conf):
    testcase = configure_testcase(file_util.parse_json(filename), conf.as_dict())

    configure_logging(testcase)

    # TODO: should be in a config somewhere
    base_url = testcase.get("base_url", "")
    url = testcase.get("url")
    brace_expansion = testcase.get("url_brace_expansion", False) or testcase.get("brace_expansion", False)
    auto_coerce = testcase.get("auto_coerce", True)
    url = urljoin(base_url, url)
    method = testcase.get("method", "get").lower()
    expected_status = testcase.get("status", 200)
    expected_response = testcase.get("response", {})
    data = testcase.get("body", None)
    upload = testcase.get("upload")
    json_encode_body = testcase.get("json_body", True)
    headers = testcase.get("headers", {})
    content_type = testcase.get("content_type", "application/json")
    headers_to_write = testcase.get("write_headers", {})
    headers_to_read = testcase.get("read_headers", {})
    match_subsets = testcase.get("match_subsets", False)
    match_falsiness = testcase.get("match_falsiness", True)

    match_options = {"match_subsets": match_subsets, "match_falsiness": match_falsiness}

    deprecation_warnings(testcase)

    if headers_to_read:
        headers = override_default_headers(headers, json.load(open(headers_to_read, "r")))

    if data:
        headers = override_default_headers(headers, {"Content-Type": content_type})

    if brace_expansion:
        brace_expander = partial(matchers.brace_expand, auto_coerce=auto_coerce)
    else:
        brace_expander = matchers.brace_expand_noop

    # we expand potential braces in the url... (eg example.com/<replace_me>/)
    url = brace_expander(url)
    # ... and each value in the dict
    data = dict_util.map_nested_dicts_py(data, brace_expander)
    # ... and also in the headers
    headers = dict_util.map_nested_dicts_py(headers, brace_expander)

    if json_encode_body:
        data = json.dumps(data)

    file = handle_upload_file(upload)

    r = http_util.do_request(url, method, data, file, headers, _logger)
    status, json_response, headers_response = r.status_code, http_util.as_json(r), r.headers

    if headers_to_write:
        dump_response_headers(headers_to_write, r)

    try:
        verify(expected_status, status, **match_options)
        verify(expected_response, json_response, **match_options)
    except Exception as e:
        error_context = {"expected": expected_response, "actual": json_response, "exception": e}
        status = STATUS_FAILED
        return status, error_context

    return "OK", None  # Yay! it passed.... nothing more to say than that


def handle_upload_file(file):
    if not file:
        return None

    key = list(file.keys())[0]
    filename = open(list(file.values())[0], 'rb')
    return {key: filename}


def dump_response_headers(headers_to_write, r):
    for filename in headers_to_write.keys():
        _logger.debug("writing header: %s" % filename)
        headers = dict_util.subset(r.headers, headers_to_write.get(filename, []))
        file_util.write_tmp(filename, json.dumps(headers))


def run():
    arguments = docopt(__doc__, version=f'skivvy {version}')
    conf = read_config(arguments.get("<cfg_file>"))
    tests = file_util.list_files(conf.tests, conf.ext)
    custom_matchers.load(conf)
    matchers.add_negating_matchers()
    fail_fast = conf.get("fail_fast", False)

    failures = 0
    num_tests = 0

    # include files - by inclusive filtering files that match the -i regexps
    # (default is ['.*'] so all files would be included in the filter)
    incl_patterns = arguments.get("-i") or []
    incl_patterns = str_util.compile_regexps(incl_patterns)
    tests = [testfile for testfile in tests if str_util.matches_any(testfile, incl_patterns)]

    # exclude files - by removing any files that match the -i regexps (default is [] so no files would be excluded)
    excl_patterns = arguments.get("-e") or []
    excl_patterns = str_util.compile_regexps(excl_patterns)
    tests = [testfile for testfile in tests if not str_util.matches_any(testfile, excl_patterns)]

    for testfile in tests:
        result, err_context = run_test(testfile, conf)
        if result == STATUS_FAILED:
            log_testcase_failed(testfile, conf)
            log_error_context(err_context, conf)
            failures += 1
        else:
            _logger.info("%s\t%s" % (testfile, STATUS_OK))
        num_tests += 1
        if fail_fast and failures > 0:
            _logger.info('Halting test run! ("fail_fast" is set to true)')
            break

    if not arguments.get("-t"):
        _logger.debug("Removing temporary files...")
        file_util.cleanup_tmp_files()

    return summary(failures, num_tests)


def summary(failures, num_tests):
    if failures > 0:
        _logger.info("%s testcases of %s failed. :(" % (failures, num_tests))
        return False
    elif num_tests == 0:
        _logger.info("No tests found!")
        return False
    else:
        _logger.info("All %s tests passed." % num_tests)
        _logger.info("Lookin' good!")
        return True


def run_skivvy():
    result = run()
    if not result:
        exit(1)
    else:
        exit(0)


if __name__ == "__main__":
    run_skivvy()



import logging

import sys


def get_logger(name, level=logging.INFO):
    # let's just use one handler for now
    # if we use a stderr handler a message will sent to both
    # pipes when an error code occurs
    # this is because there's no way to set a range
    # for a handler's log level. Eg "emit only debug & info"

    # anyway, when we had a stderr handler text
    # was duplicated in the terminal in an annoying way
    # there's probably some way around this that I haven't found,,,

    logger = logging.getLogger(name)
    default_handler = logging.StreamHandler(sys.stdout)
    default_handler.setLevel(logging.DEBUG)
    logger.addHandler(default_handler)
    logger.setLevel(level)

    # error_handler = logging.StreamHandler(sys.stderr)
    # error_handler.setLevel(logging.WARNING)
    # logger.addHandler(error_handler)
    return logger
